@@ -24,7 +24,7 @@
 const util = require('util');
 const net = require('net');
 const url = require('url');
-const { HTTPParser } = process.binding('http_parser');
+const { HTTPParser } = internalBinding('http_parser');
 const assert = require('assert').ok;
 const {
   _checkIsHttpToken: checkIsHttpToken,
@@ -36,62 +36,56 @@ const {
 const { OutgoingMessage } = require('_http_outgoing');
 const Agent = require('_http_agent');
 const { Buffer } = require('buffer');
+const { defaultTriggerAsyncIdScope } = require('internal/async_hooks');
 const { urlToOptions, searchParamsSymbol } = require('internal/url');
 const { outHeadersKey, ondrain } = require('internal/http');
-const { nextTick } = require('internal/process/next_tick');
+const {
+  ERR_HTTP_HEADERS_SENT,
+  ERR_INVALID_ARG_TYPE,
+  ERR_INVALID_DOMAIN_NAME,
+  ERR_INVALID_HTTP_TOKEN,
+  ERR_INVALID_PROTOCOL,
+  ERR_UNESCAPED_CHARACTERS
+} = require('internal/errors').codes;
+const { validateTimerDuration } = require('internal/timers');
+const is_reused_symbol = require('internal/freelist').symbols.is_reused_symbol;
 
-// The actual list of disallowed characters in regexp form is more like:
-//    /[^A-Za-z0-9\-._~!$&'()*+,;=/:@]/
-// with an additional rule for ignoring percentage-escaped characters, but
-// that's a) hard to capture in a regular expression that performs well, and
-// b) possibly too restrictive for real-world usage. So instead we restrict the
-// filter to just control characters and spaces.
-//
-// This function is used in the case of small paths, where manual character code
-// checks can greatly outperform the equivalent regexp (tested in V8 5.4).
-function isInvalidPath(s) {
-  var i = 0;
-  if (s.charCodeAt(0) <= 32) return true;
-  if (++i >= s.length) return false;
-  if (s.charCodeAt(1) <= 32) return true;
-  if (++i >= s.length) return false;
-  if (s.charCodeAt(2) <= 32) return true;
-  if (++i >= s.length) return false;
-  if (s.charCodeAt(3) <= 32) return true;
-  if (++i >= s.length) return false;
-  if (s.charCodeAt(4) <= 32) return true;
-  if (++i >= s.length) return false;
-  if (s.charCodeAt(5) <= 32) return true;
-  ++i;
-  for (; i < s.length; ++i)
-    if (s.charCodeAt(i) <= 32) return true;
-  return false;
-}
+const INVALID_PATH_REGEX = /[^\u0021-\u00ff]/;
 
 function validateHost(host, name) {
-  if (host != null && typeof host !== 'string') {
-    throw new TypeError(
-      `"options.${name}" must either be a string, undefined or null`);
+  if (host !== null && host !== undefined && typeof host !== 'string') {
+    throw new ERR_INVALID_ARG_TYPE(`options.${name}`,
+                                   ['string', 'undefined', 'null'],
+                                   host);
   }
   return host;
 }
 
-function ClientRequest(options, cb) {
+function ClientRequest(input, options, cb) {
   OutgoingMessage.call(this);
 
-  if (typeof options === 'string') {
-    options = url.parse(options);
-    if (!options.hostname) {
-      throw new Error('Unable to determine the domain name');
+  if (typeof input === 'string') {
+    input = url.parse(input);
+    if (!input.hostname) {
+      throw new ERR_INVALID_DOMAIN_NAME();
     }
-  } else if (options && options[searchParamsSymbol] &&
-             options[searchParamsSymbol][searchParamsSymbol]) {
+  } else if (input && input[searchParamsSymbol] &&
+             input[searchParamsSymbol][searchParamsSymbol]) {
     // url.URL instance
-    options = urlToOptions(options);
+    input = urlToOptions(input);
   } else {
-    options = util._extend({}, options);
+    cb = options;
+    options = input;
+    input = null;
   }
 
+  if (typeof options === 'function') {
+    cb = options;
+    options = null;
+  }
+
+  options = util._extend(input || {}, options || {});
+
   var agent = options.agent;
   var defaultAgent = options._defaultAgent || Agent.globalAgent;
   if (agent === false) {
@@ -103,9 +97,9 @@ function ClientRequest(options, cb) {
     // Explicitly pass through this statement as agent will not be used
     // when createConnection is provided.
   } else if (typeof agent.addRequest !== 'function') {
-    throw new TypeError(
-      'Agent option must be an Agent-like object, undefined, or false.'
-    );
+    throw new ERR_INVALID_ARG_TYPE('options.agent',
+                                   ['Agent-like Object', 'undefined', 'false'],
+                                   agent);
   }
   this.agent = agent;
 
@@ -117,19 +111,12 @@ function ClientRequest(options, cb) {
   var path;
   if (options.path) {
     path = String(options.path);
-    var invalidPath;
-    if (path.length <= 39) { // Determined experimentally in V8 5.4
-      invalidPath = isInvalidPath(path);
-    } else {
-      invalidPath = /[\u0000-\u0020]/.test(path);
-    }
-    if (invalidPath)
-      throw new TypeError('Request path contains unescaped characters');
+    if (INVALID_PATH_REGEX.test(path))
+      throw new ERR_UNESCAPED_CHARACTERS('Request path');
   }
 
   if (protocol !== expectedProtocol) {
-    throw new Error('Protocol "' + protocol + '" not supported. ' +
-                    'Expected "' + expectedProtocol + '"');
+    throw new ERR_INVALID_PROTOCOL(protocol, expectedProtocol);
   }
 
   var defaultPort = options.defaultPort ||
@@ -139,20 +126,20 @@ function ClientRequest(options, cb) {
   var host = options.host = validateHost(options.hostname, 'hostname') ||
                             validateHost(options.host, 'host') || 'localhost';
 
-  var setHost = (options.setHost === undefined);
+  var setHost = (options.setHost === undefined || Boolean(options.setHost));
 
   this.socketPath = options.socketPath;
   this.timeout = options.timeout;
 
   var method = options.method;
   var methodIsString = (typeof method === 'string');
-  if (method != null && !methodIsString) {
-    throw new TypeError('Method must be a string');
+  if (method !== null && method !== undefined && !methodIsString) {
+    throw new ERR_INVALID_ARG_TYPE('method', 'string', method);
   }
 
   if (methodIsString && method) {
     if (!checkIsHttpToken(method)) {
-      throw new TypeError('Method must be a valid HTTP token');
+      throw new ERR_INVALID_HTTP_TOKEN('Method', method);
     }
     method = this.method = method.toUpperCase();
   } else {
@@ -164,40 +151,6 @@ function ClientRequest(options, cb) {
     this.once('response', cb);
   }
 
-  var headersArray = Array.isArray(options.headers);
-  if (!headersArray) {
-    if (options.headers) {
-      var keys = Object.keys(options.headers);
-      for (var i = 0; i < keys.length; i++) {
-        var key = keys[i];
-        this.setHeader(key, options.headers[key]);
-      }
-    }
-    if (host && !this.getHeader('host') && setHost) {
-      var hostHeader = host;
-
-      // For the Host header, ensure that IPv6 addresses are enclosed
-      // in square brackets, as defined by URI formatting
-      // https://tools.ietf.org/html/rfc3986#section-3.2.2
-      var posColon = hostHeader.indexOf(':');
-      if (posColon !== -1 &&
-          hostHeader.indexOf(':', posColon + 1) !== -1 &&
-          hostHeader.charCodeAt(0) !== 91/*'['*/) {
-        hostHeader = `[${hostHeader}]`;
-      }
-
-      if (port && +port !== defaultPort) {
-        hostHeader += ':' + port;
-      }
-      this.setHeader('Host', hostHeader);
-    }
-  }
-
-  if (options.auth && !this.getHeader('Authorization')) {
-    this.setHeader('Authorization', 'Basic ' +
-                   Buffer.from(options.auth).toString('base64'));
-  }
-
   if (method === 'GET' ||
       method === 'HEAD' ||
       method === 'DELETE' ||
@@ -208,19 +161,6 @@ function ClientRequest(options, cb) {
     this.useChunkedEncodingByDefault = true;
   }
 
-  if (headersArray) {
-    this._storeHeader(this.method + ' ' + this.path + ' HTTP/1.1\r\n',
-                      options.headers);
-  } else if (this.getHeader('expect')) {
-    if (this._header) {
-      throw new Error('Can\'t render headers after they are sent to the ' +
-                      'client');
-    }
-
-    this._storeHeader(this.method + ' ' + this.path + ' HTTP/1.1\r\n',
-                      this[outHeadersKey]);
-  }
-
   this._ended = false;
   this.res = null;
   this.aborted = undefined;
@@ -231,18 +171,6 @@ function ClientRequest(options, cb) {
 
   var called = false;
 
-  var oncreate = (err, socket) => {
-    if (called)
-      return;
-    called = true;
-    if (err) {
-      process.nextTick(() => this.emit('error', err));
-      return;
-    }
-    this.onSocket(socket);
-    this._deferToConnect(null, null, () => this._flush());
-  };
-
   if (this.agent) {
     // If there is an agent we should default to Connection:keep-alive,
     // but only if the Agent will actually reuse the connection!
@@ -255,6 +183,69 @@ function ClientRequest(options, cb) {
       this._last = false;
       this.shouldKeepAlive = true;
     }
+  }
+
+  var headersArray = Array.isArray(options.headers);
+  if (!headersArray) {
+    if (options.headers) {
+      var keys = Object.keys(options.headers);
+      for (var i = 0; i < keys.length; i++) {
+        var key = keys[i];
+        this.setHeader(key, options.headers[key]);
+      }
+    }
+
+    if (host && !this.getHeader('host') && setHost) {
+      var hostHeader = host;
+
+      // For the Host header, ensure that IPv6 addresses are enclosed
+      // in square brackets, as defined by URI formatting
+      // https://tools.ietf.org/html/rfc3986#section-3.2.2
+      var posColon = hostHeader.indexOf(':');
+      if (posColon !== -1 &&
+          hostHeader.indexOf(':', posColon + 1) !== -1 &&
+          hostHeader.charCodeAt(0) !== 91/* '[' */) {
+        hostHeader = `[${hostHeader}]`;
+      }
+
+      if (port && +port !== defaultPort) {
+        hostHeader += ':' + port;
+      }
+      this.setHeader('Host', hostHeader);
+    }
+
+    if (options.auth && !this.getHeader('Authorization')) {
+      this.setHeader('Authorization', 'Basic ' +
+                     Buffer.from(options.auth).toString('base64'));
+    }
+
+    if (this.getHeader('expect')) {
+      if (this._header) {
+        throw new ERR_HTTP_HEADERS_SENT('render');
+      }
+
+      this._storeHeader(this.method + ' ' + this.path + ' HTTP/1.1\r\n',
+                        this[outHeadersKey]);
+    }
+  } else {
+    this._storeHeader(this.method + ' ' + this.path + ' HTTP/1.1\r\n',
+                      options.headers);
+  }
+
+  var oncreate = (err, socket) => {
+    if (called)
+      return;
+    called = true;
+    if (err) {
+      process.nextTick(() => this.emit('error', err));
+      return;
+    }
+    this.onSocket(socket);
+    this._deferToConnect(null, null, () => this._flush());
+  };
+
+  // initiate connection
+  if (this.agent) {
     this.agent.addRequest(this, options);
   } else {
     // No agent, default to Connection:close.
@@ -282,14 +273,13 @@ util.inherits(ClientRequest, OutgoingMessage);
 
 ClientRequest.prototype._finish = function _finish() {
   DTRACE_HTTP_CLIENT_REQUEST(this, this.connection);
-  LTTNG_HTTP_CLIENT_REQUEST(this, this.connection);
   COUNTER_HTTP_CLIENT_REQUEST();
   OutgoingMessage.prototype._finish.call(this);
 };
 
 ClientRequest.prototype._implicitHeader = function _implicitHeader() {
   if (this._header) {
-    throw new Error('Can\'t render headers after they are sent to the client');
+    throw new ERR_HTTP_HEADERS_SENT('render');
   }
   this._storeHeader(this.method + ' ' + this.path + ' HTTP/1.1\r\n',
                     this[outHeadersKey]);
@@ -299,6 +289,7 @@ ClientRequest.prototype.abort = function abort() {
   if (!this.aborted) {
     process.nextTick(emitAbortNT.bind(this));
   }
+
   // Mark as aborting so we can avoid sending queued request data
   // This is used as a truthy flag elsewhere. The use of Date.now is for
   // debugging purposes only.
@@ -328,6 +319,7 @@ function emitAbortNT() {
 
 
 function createHangUpError() {
+  // eslint-disable-next-line no-restricted-syntax
   var error = new Error('socket hang up');
   error.code = 'ECONNRESET';
   return error;
@@ -346,22 +338,32 @@ function socketCloseListener() {
 
   // NOTE: It's important to get parser here, because it could be freed by
   // the `socketOnData`.
-  var parser = socket.parser;
-  req.emit('close');
-  if (req.res && req.res.readable) {
+  const parser = socket.parser;
+  const res = req.res;
+  if (res) {
     // Socket closed before we emitted 'end' below.
-    if (!req.res.complete) req.res.emit('aborted');
-    var res = req.res;
-    res.on('end', function() {
+    if (!res.complete) {
+      res.aborted = true;
+      res.emit('aborted');
+    }
+    req.emit('close');
+    if (res.readable) {
+      res.on('end', function() {
+        this.emit('close');
+      });
+      res.push(null);
+    } else {
       res.emit('close');
-    });
-    res.push(null);
-  } else if (!req.res && !req.socket._hadError) {
-    // This socket error fired before we started to
-    // receive a response. The error needs to
-    // fire on the request.
-    req.emit('error', createHangUpError());
-    req.socket._hadError = true;
+    }
+  } else {
+    if (!req.socket._hadError) {
+      // This socket error fired before we started to
+      // receive a response. The error needs to
+      // fire on the request.
+      req.socket._hadError = true;
+      req.emit('error', createHangUpError());
+    }
+    req.emit('close');
   }
 
   // Too bad.  That output wasn't getting written.
@@ -384,10 +386,10 @@ function socketErrorListener(err) {
   debug('SOCKET ERROR:', err.message, err.stack);
 
   if (req) {
-    req.emit('error', err);
     // For Safety. Some additional errors might fire later on
     // and we need to make sure we don't double-fire the error event.
     req.socket._hadError = true;
+    req.emit('error', err);
   }
 
   // Handle any pending data
@@ -420,8 +422,8 @@ function socketOnEnd() {
   if (!req.res && !req.socket._hadError) {
     // If we don't have a response then we know that the socket
     // ended prematurely and we need to emit an error on the request.
-    req.emit('error', createHangUpError());
     req.socket._hadError = true;
+    req.emit('error', createHangUpError());
   }
   if (parser) {
     parser.finish();
@@ -442,10 +444,10 @@ function socketOnData(d) {
     debug('parse error', ret);
     freeParser(parser, req, socket);
     socket.destroy();
-    req.emit('error', ret);
     req.socket._hadError = true;
+    req.emit('error', ret);
   } else if (parser.incoming && parser.incoming.upgrade) {
-    // Upgrade or CONNECT
+    // Upgrade (if status code 101) or CONNECT
     var bytesParsed = ret;
     var res = parser.incoming;
     req.res = res;
@@ -467,40 +469,41 @@ function socketOnData(d) {
       socket.removeListener('close', socketCloseListener);
       socket.removeListener('error', socketErrorListener);
 
-      // TODO(isaacs): Need a way to reset a stream to fresh state
-      // IE, not flowing, and not explicitly paused.
-      socket._readableState.flowing = null;
       socket._httpMessage = null;
+      socket.readableFlowing = null;
 
       req.emit(eventName, res, socket, bodyHead);
       req.emit('close');
     } else {
-      // Got Upgrade header or CONNECT method, but have no handler.
+      // Requested Upgrade or used CONNECT method, but have no handler.
       socket.destroy();
     }
   } else if (parser.incoming && parser.incoming.complete &&
-             // When the status code is 100 (Continue), the server will
-             // send a final response after this client sends a request
-             // body. So, we must not free the parser.
-             parser.incoming.statusCode !== 100) {
+             // When the status code is informational (100, 102-199),
+             // the server will send a final response after this client
+             // sends a request body, so we must not free the parser.
+             // 101 (Switching Protocols) and all other status codes
+             // should be processed normally.
+             !statusIsInformational(parser.incoming.statusCode)) {
     socket.removeListener('data', socketOnData);
     socket.removeListener('end', socketOnEnd);
     freeParser(parser, req, socket);
   }
 }
 
+function statusIsInformational(status) {
+  // 100 (Continue)    RFC7231 Section 6.2.1
+  // 102 (Processing)  RFC2518
+  // 103 (Early Hints) RFC8297
+  // 104-199 (Unassigned)
+  return (status < 200 && status >= 100 && status !== 101);
+}
 
 // client
 function parserOnIncomingClient(res, shouldKeepAlive) {
   var socket = this.socket;
   var req = socket._httpMessage;
 
-  // propagate "domain" setting...
-  if (req.domain && !res.domain) {
-    debug('setting "res.domain"');
-    res.domain = req.domain;
-  }
-
   debug('AGENT incoming response!');
 
   if (req.res) {
@@ -511,6 +514,10 @@ function parserOnIncomingClient(res, shouldKeepAlive) {
   }
   req.res = res;
 
+  // Skip body and treat as Upgrade.
+  if (res.upgrade)
+    return 2;
+
   // Responses to CONNECT request is handled as Upgrade.
   const method = req.method;
   if (method === 'CONNECT') {
@@ -518,10 +525,16 @@ function parserOnIncomingClient(res, shouldKeepAlive) {
     return 2;  // Skip body and treat as Upgrade.
   }
 
-  if (res.statusCode === 100) {
-    // restart the parser, as this is a continue message.
+  if (statusIsInformational(res.statusCode)) {
+    // Restart the parser, as this is a 1xx informational message.
     req.res = null; // Clear res so that we don't hit double-responses.
-    req.emit('continue');
+    // Maintain compatibility by sending 100-specific events
+    if (res.statusCode === 100) {
+      req.emit('continue');
+    }
+    // Send information events to all 1xx responses except 101 Upgrade.
+    req.emit('information', { statusCode: res.statusCode });
+
     return 1;  // Skip body but don't treat as Upgrade.
   }
 
@@ -533,7 +546,6 @@ function parserOnIncomingClient(res, shouldKeepAlive) {
   }
 
   DTRACE_HTTP_CLIENT_RESPONSE(socket, req);
-  LTTNG_HTTP_CLIENT_RESPONSE(socket, req);
   COUNTER_HTTP_CLIENT_RESPONSE();
   req.res = res;
   res.req = req;
@@ -579,10 +591,10 @@ function responseKeepAlive(res, req) {
     socket.once('error', freeSocketErrorListener);
     // There are cases where _handle === null. Avoid those. Passing null to
     // nextTick() will call getDefaultTriggerAsyncId() to retrieve the id.
-    const asyncId = socket._handle ? socket._handle.getAsyncId() : null;
+    const asyncId = socket._handle ? socket._handle.getAsyncId() : undefined;
     // Mark this socket as available, AFTER user-added end
     // handlers have a chance to run.
-    nextTick(asyncId, emitFreeNT, socket);
+    defaultTriggerAsyncIdScope(asyncId, process.nextTick, emitFreeNT, socket);
   }
 }
 
@@ -614,9 +626,8 @@ function tickOnSocket(req, socket) {
   var parser = parsers.alloc();
   req.socket = socket;
   req.connection = socket;
-  parser.reinitialize(HTTPParser.RESPONSE);
+  parser.reinitialize(HTTPParser.RESPONSE, parser[is_reused_symbol]);
   parser.socket = socket;
-  parser.incoming = null;
   parser.outgoing = req;
   req.parser = parser;
 
@@ -629,9 +640,6 @@ function tickOnSocket(req, socket) {
   // Propagate headers limit from request object to parser
   if (typeof req.maxHeadersCount === 'number') {
     parser.maxHeaderPairs = req.maxHeadersCount << 1;
-  } else {
-    // Set default value because parser may be reused from FreeList
-    parser.maxHeaderPairs = 2000;
   }
 
   parser.onIncoming = parserOnIncomingClient;
@@ -641,18 +649,35 @@ function tickOnSocket(req, socket) {
   socket.on('end', socketOnEnd);
   socket.on('close', socketCloseListener);
 
-  if (req.timeout) {
-    const emitRequestTimeout = () => req.emit('timeout');
-    socket.once('timeout', emitRequestTimeout);
-    req.once('response', (res) => {
-      res.once('end', () => {
-        socket.removeListener('timeout', emitRequestTimeout);
-      });
-    });
+  if (req.timeout !== undefined) {
+    listenSocketTimeout(req);
   }
   req.emit('socket', socket);
 }
 
+function listenSocketTimeout(req) {
+  if (req.timeoutCb) {
+    return;
+  }
+  const emitRequestTimeout = () => req.emit('timeout');
+  // Set timeoutCb so it will get cleaned up on request end.
+  req.timeoutCb = emitRequestTimeout;
+  // Delegate socket timeout event.
+  if (req.socket) {
+    req.socket.once('timeout', emitRequestTimeout);
+  } else {
+    req.on('socket', (socket) => {
+      socket.once('timeout', emitRequestTimeout);
+    });
+  }
+  // Remove socket timeout listener after response end.
+  req.once('response', (res) => {
+    res.once('end', () => {
+      req.socket.removeListener('timeout', emitRequestTimeout);
+    });
+  });
+}
+
 ClientRequest.prototype.onSocket = function onSocket(socket) {
   process.nextTick(onSocketNT, this, socket);
 };
@@ -702,35 +727,29 @@ function _deferToConnect(method, arguments_, cb) {
 }
 
 ClientRequest.prototype.setTimeout = function setTimeout(msecs, callback) {
+  listenSocketTimeout(this);
+  msecs = validateTimerDuration(msecs);
   if (callback) this.once('timeout', callback);
 
-  const emitTimeout = () => this.emit('timeout');
-
-  if (this.socket && this.socket.writable) {
-    if (this.timeoutCb)
-      this.socket.setTimeout(0, this.timeoutCb);
-    this.timeoutCb = emitTimeout;
-    this.socket.setTimeout(msecs, emitTimeout);
-    return this;
-  }
-
-  // Set timeoutCb so that it'll get cleaned up on request end
-  this.timeoutCb = emitTimeout;
   if (this.socket) {
-    var sock = this.socket;
-    this.socket.once('connect', function() {
-      sock.setTimeout(msecs, emitTimeout);
-    });
-    return this;
+    setSocketTimeout(this.socket, msecs);
+  } else {
+    this.once('socket', (sock) => setSocketTimeout(sock, msecs));
   }
 
-  this.once('socket', function(sock) {
-    sock.setTimeout(msecs, emitTimeout);
-  });
-
   return this;
 };
 
+function setSocketTimeout(sock, msecs) {
+  if (sock.connecting) {
+    sock.once('connect', function() {
+      sock.setTimeout(msecs);
+    });
+  } else {
+    sock.setTimeout(msecs);
+  }
+}
+
 ClientRequest.prototype.setNoDelay = function setNoDelay(noDelay) {
   this._deferToConnect('setNoDelay', [noDelay]);
 };
