@@ -1,33 +1,58 @@
 'use strict';
 
-const StringDecoder = require('string_decoder').StringDecoder;
-const Buffer = require('buffer').Buffer;
+const {
+  errnoException,
+  codes: {
+    ERR_INVALID_ARG_TYPE,
+    ERR_INVALID_HANDLE_TYPE,
+    ERR_INVALID_OPT_VALUE,
+    ERR_INVALID_SYNC_FORK_INPUT,
+    ERR_IPC_CHANNEL_CLOSED,
+    ERR_IPC_DISCONNECTED,
+    ERR_IPC_ONE_PIPE,
+    ERR_IPC_SYNC_FORK,
+    ERR_MISSING_ARGS
+  }
+} = require('internal/errors');
+const { validateString } = require('internal/validators');
 const EventEmitter = require('events');
 const net = require('net');
 const dgram = require('dgram');
 const util = require('util');
-const constants = process.binding('constants').os.signals;
 const assert = require('assert');
 
-const Process = process.binding('process_wrap').Process;
-const WriteWrap = process.binding('stream_wrap').WriteWrap;
-const uv = process.binding('uv');
-const Pipe = process.binding('pipe_wrap').Pipe;
-const TTY = process.binding('tty_wrap').TTY;
-const TCP = process.binding('tcp_wrap').TCP;
-const UDP = process.binding('udp_wrap').UDP;
+const { Process } = internalBinding('process_wrap');
+const { WriteWrap } = internalBinding('stream_wrap');
+const { Pipe, constants: PipeConstants } = internalBinding('pipe_wrap');
+const { TCP } = internalBinding('tcp_wrap');
+const { TTY } = internalBinding('tty_wrap');
+const { UDP } = internalBinding('udp_wrap');
 const SocketList = require('internal/socket_list');
+const { owner_symbol } = require('internal/async_hooks').symbols;
+const { convertToValidSignal } = require('internal/util');
+const { isArrayBufferView } = require('internal/util/types');
+const spawn_sync = process.binding('spawn_sync');
+const { HTTPParser } = process.binding('http_parser');
+const { freeParser } = require('_http_common');
+const { kStateSymbol } = require('internal/dgram');
 
-const errnoException = util._errnoException;
-const SocketListSend = SocketList.SocketListSend;
-const SocketListReceive = SocketList.SocketListReceive;
+const {
+  UV_EACCES,
+  UV_EAGAIN,
+  UV_EINVAL,
+  UV_EMFILE,
+  UV_ENFILE,
+  UV_ENOENT,
+  UV_ENOSYS,
+  UV_ESRCH
+} = process.binding('uv');
 
-module.exports = {
-  ChildProcess,
-  setupChannel,
-  _validateStdio,
-  getSocketList
-};
+const { SocketListSend, SocketListReceive } = SocketList;
+
+// Lazy loaded for startup performance.
+let StringDecoder;
+
+const MAX_HANDLE_RETRANSMISSIONS = 3;
 
 // this object contain function to convert TCP objects to native handle objects
 // and back again.
@@ -66,16 +91,16 @@ const handleConversion = {
 
       // if the socket was created by net.Server
       if (socket.server) {
-        // the slave should keep track of the socket
+        // the worker should keep track of the socket
         message.key = socket.server._connectionKey;
 
-        var firstTime = !this._channel.sockets.send[message.key];
+        var firstTime = !this.channel.sockets.send[message.key];
         var socketList = getSocketList('send', this, message.key);
 
         // the server should no longer expose a .connection property
         // and when asked to close it should query the socket status from
-        // the slaves
-        if (firstTime) socket.server._setupSlave(socketList);
+        // the workers
+        if (firstTime) socket.server._setupWorker(socketList);
 
         // Act like socket is detached
         if (!options.keepOpen)
@@ -89,22 +114,31 @@ const handleConversion = {
       if (!options.keepOpen) {
         handle.onread = nop;
         socket._handle = null;
+        socket.setTimeout(0);
+        // In case of an HTTP connection socket, release the associated
+        // resources
+        if (socket.parser && socket.parser instanceof HTTPParser) {
+          freeParser(socket.parser, null, socket);
+          if (socket._httpMessage)
+            socket._httpMessage.detachSocket(socket);
+        }
       }
 
       return handle;
     },
 
-    postSend: function(handle, options, target) {
+    postSend: function(message, handle, options, callback, target) {
       // Store the handle after successfully sending it, so it can be closed
       // when the NODE_HANDLE_ACK is received. If the handle could not be sent,
       // just close it.
       if (handle && !options.keepOpen) {
         if (target) {
-          // There can only be one _pendingHandle as passing handles are
+          // There can only be one _pendingMessage as passing handles are
           // processed one at a time: handles are stored in _handleQueue while
           // waiting for the NODE_HANDLE_ACK of the current passing handle.
-          assert(!target._pendingHandle);
-          target._pendingHandle = handle;
+          assert(!target._pendingMessage);
+          target._pendingMessage =
+              { callback, message, handle, options, retransmissions: 0 };
         } else {
           handle.close();
         }
@@ -112,8 +146,11 @@ const handleConversion = {
     },
 
     got: function(message, handle, emit) {
-      var socket = new net.Socket({handle: handle});
-      socket.readable = socket.writable = true;
+      var socket = new net.Socket({
+        handle: handle,
+        readable: true,
+        writable: true
+      });
 
       // if the socket was created by net.Server we will track the socket
       if (message.key) {
@@ -147,7 +184,7 @@ const handleConversion = {
     send: function(message, socket, options) {
       message.dgramType = socket.type;
 
-      return socket._handle;
+      return socket[kStateSymbol].handle;
     },
 
     got: function(message, handle, emit) {
@@ -164,8 +201,6 @@ const handleConversion = {
 function ChildProcess() {
   EventEmitter.call(this);
 
-  var self = this;
-
   this._closesNeeded = 1;
   this._closesGot = 0;
   this.connected = false;
@@ -176,43 +211,33 @@ function ChildProcess() {
   this.spawnfile = null;
 
   this._handle = new Process();
-  this._handle.owner = this;
-
-  this._handle.onexit = function(exitCode, signalCode) {
-    //
-    // follow 0.4.x behaviour:
-    //
-    // - normally terminated processes don't touch this.signalCode
-    // - signaled processes don't touch this.exitCode
-    //
-    // new in 0.9.x:
-    //
-    // - spawn failures are reported with exitCode < 0
-    //
-    var syscall = self.spawnfile ? 'spawn ' + self.spawnfile : 'spawn';
-    var err = (exitCode < 0) ? errnoException(exitCode, syscall) : null;
+  this._handle[owner_symbol] = this;
 
+  this._handle.onexit = (exitCode, signalCode) => {
     if (signalCode) {
-      self.signalCode = signalCode;
+      this.signalCode = signalCode;
     } else {
-      self.exitCode = exitCode;
+      this.exitCode = exitCode;
     }
 
-    if (self.stdin) {
-      self.stdin.destroy();
+    if (this.stdin) {
+      this.stdin.destroy();
     }
 
-    self._handle.close();
-    self._handle = null;
+    this._handle.close();
+    this._handle = null;
 
     if (exitCode < 0) {
-      if (self.spawnfile)
-        err.path = self.spawnfile;
+      var syscall = this.spawnfile ? 'spawn ' + this.spawnfile : 'spawn';
+      const err = errnoException(exitCode, syscall);
 
-      err.spawnargs = self.spawnargs.slice(1);
-      self.emit('error', err);
+      if (this.spawnfile)
+        err.path = this.spawnfile;
+
+      err.spawnargs = this.spawnargs.slice(1);
+      this.emit('error', err);
     } else {
-      self.emit('exit', self.exitCode, self.signalCode);
+      this.emit('exit', this.exitCode, this.signalCode);
     }
 
     // if any of the stdio streams have not been touched,
@@ -221,9 +246,9 @@ function ChildProcess() {
     // Do it on nextTick so that the user has one last chance
     // to consume the output, if for example they only want to
     // start reading the data once the process exits.
-    process.nextTick(flushStdio, self);
+    process.nextTick(flushStdio, this);
 
-    maybeClose(self);
+    maybeClose(this);
   };
 }
 util.inherits(ChildProcess, EventEmitter);
@@ -244,17 +269,7 @@ function flushStdio(subprocess) {
 
 
 function createSocket(pipe, readable) {
-  var s = new net.Socket({ handle: pipe });
-
-  if (readable) {
-    s.writable = false;
-    s.readable = true;
-  } else {
-    s.writable = true;
-    s.readable = false;
-  }
-
-  return s;
+  return net.Socket({ handle: pipe, readable, writable: !readable });
 }
 
 
@@ -267,12 +282,21 @@ function getHandleWrapType(stream) {
   return false;
 }
 
+function closePendingHandle(target) {
+  target._pendingMessage.handle.close();
+  target._pendingMessage = null;
+}
+
 
 ChildProcess.prototype.spawn = function(options) {
-  const self = this;
   var ipc;
   var ipcFd;
   var i;
+
+  if (options === null || typeof options !== 'object') {
+    throw new ERR_INVALID_ARG_TYPE('options', 'Object', options);
+  }
+
   // If no `stdio` option was given - use default
   var stdio = options.stdio || 'pipe';
 
@@ -284,27 +308,42 @@ ChildProcess.prototype.spawn = function(options) {
 
   if (ipc !== undefined) {
     // Let child process know about opened IPC channel
-    options.envPairs = options.envPairs || [];
+    if (options.envPairs === undefined)
+      options.envPairs = [];
+    else if (!Array.isArray(options.envPairs)) {
+      throw new ERR_INVALID_ARG_TYPE('options.envPairs',
+                                     'Array',
+                                     options.envPairs);
+    }
+
     options.envPairs.push('NODE_CHANNEL_FD=' + ipcFd);
   }
 
+  validateString(options.file, 'options.file');
   this.spawnfile = options.file;
-  this.spawnargs = options.args;
+
+  if (Array.isArray(options.args))
+    this.spawnargs = options.args;
+  else if (options.args === undefined)
+    this.spawnargs = [];
+  else
+    throw new ERR_INVALID_ARG_TYPE('options.args', 'Array', options.args);
 
   var err = this._handle.spawn(options);
 
   // Run-time errors should emit an error, not throw an exception.
-  if (err === uv.UV_EAGAIN ||
-      err === uv.UV_EMFILE ||
-      err === uv.UV_ENFILE ||
-      err === uv.UV_ENOENT) {
-    process.nextTick(onErrorNT, self, err);
+  if (err === UV_EACCES ||
+      err === UV_EAGAIN ||
+      err === UV_EMFILE ||
+      err === UV_ENFILE ||
+      err === UV_ENOENT) {
+    process.nextTick(onErrorNT, this, err);
     // There is no point in continuing when we've hit EMFILE or ENFILE
     // because we won't be able to set up the stdio file descriptors.
     // It's kind of silly that the de facto spec for ENOENT (the test suite)
     // mandates that stdio _is_ set up, even if there is no process on the
     // receiving end, but it is what it is.
-    if (err !== uv.UV_ENOENT) return err;
+    if (err !== UV_ENOENT) return err;
   } else if (err) {
     // Close all opened fds on error
     for (i = 0; i < stdio.length; i++) {
@@ -326,35 +365,36 @@ ChildProcess.prototype.spawn = function(options) {
     if (stream.type === 'ignore') continue;
 
     if (stream.ipc) {
-      self._closesNeeded++;
+      this._closesNeeded++;
       continue;
     }
 
     if (stream.handle) {
       // when i === 0 - we're dealing with stdin
       // (which is the only one writable pipe)
-      stream.socket = createSocket(self.pid !== 0 ?
-          stream.handle : null, i > 0);
+      stream.socket = createSocket(this.pid !== 0 ?
+        stream.handle : null, i > 0);
 
-      if (i > 0 && self.pid !== 0) {
-        self._closesNeeded++;
-        stream.socket.on('close', function() {
-          maybeClose(self);
+      if (i > 0 && this.pid !== 0) {
+        this._closesNeeded++;
+        stream.socket.on('close', () => {
+          maybeClose(this);
         });
       }
     }
   }
 
   this.stdin = stdio.length >= 1 && stdio[0].socket !== undefined ?
-      stdio[0].socket : null;
+    stdio[0].socket : null;
   this.stdout = stdio.length >= 2 && stdio[1].socket !== undefined ?
-      stdio[1].socket : null;
+    stdio[1].socket : null;
   this.stderr = stdio.length >= 3 && stdio[2].socket !== undefined ?
-      stdio[2].socket : null;
+    stdio[2].socket : null;
 
-  this.stdio = stdio.map(function(stdio) {
-    return stdio.socket === undefined ? null : stdio.socket;
-  });
+  this.stdio = [];
+
+  for (i = 0; i < stdio.length; i++)
+    this.stdio.push(stdio[i].socket === undefined ? null : stdio[i].socket);
 
   // Add .send() method and start listening for IPC data
   if (ipc !== undefined) setupChannel(this, ipc);
@@ -369,19 +409,9 @@ function onErrorNT(self, err) {
 
 
 ChildProcess.prototype.kill = function(sig) {
-  var signal;
 
-  if (sig === 0) {
-    signal = 0;
-  } else if (!sig) {
-    signal = constants['SIGTERM'];
-  } else {
-    signal = constants[sig];
-  }
-
-  if (signal === undefined) {
-    throw new Error('Unknown signal: ' + sig);
-  }
+  const signal = sig === 0 ? sig :
+    convertToValidSignal(sig === undefined ? 'SIGTERM' : sig);
 
   if (this._handle) {
     var err = this._handle.kill(signal);
@@ -390,9 +420,9 @@ ChildProcess.prototype.kill = function(sig) {
       this.killed = true;
       return true;
     }
-    if (err === uv.UV_ESRCH) {
+    if (err === UV_ESRCH) {
       /* Already dead. */
-    } else if (err === uv.UV_EINVAL || err === uv.UV_ENOSYS) {
+    } else if (err === UV_EINVAL || err === UV_ENOSYS) {
       /* The underlying platform doesn't support this signal. */
       throw errnoException(err, 'kill');
     } else {
@@ -435,18 +465,35 @@ class Control extends EventEmitter {
 }
 
 function setupChannel(target, channel) {
-  target._channel = channel;
+  target.channel = channel;
+
+  // _channel can be deprecated in version 8
+  Object.defineProperty(target, '_channel', {
+    get() { return target.channel; },
+    set(val) { target.channel = val; },
+    enumerable: true
+  });
+
   target._handleQueue = null;
-  target._pendingHandle = null;
+  target._pendingMessage = null;
 
   const control = new Control(channel);
 
+  if (StringDecoder === undefined)
+    StringDecoder = require('string_decoder').StringDecoder;
   var decoder = new StringDecoder('utf8');
   var jsonBuffer = '';
+  var pendingHandle = null;
   channel.buffering = false;
-  channel.onread = function(nread, pool, recvHandle) {
+  channel.pendingHandle = null;
+  channel.onread = function(nread, pool) {
+    const recvHandle = channel.pendingHandle;
+    channel.pendingHandle = null;
     // TODO(bnoordhuis) Check that nread > 0.
     if (pool) {
+      if (recvHandle)
+        pendingHandle = recvHandle;
+
       // Linebreak is used as a message end sign
       var chunks = decoder.write(pool).split('\n');
       var numCompleteChunks = chunks.length - 1;
@@ -465,10 +512,16 @@ function setupChannel(target, channel) {
         // There will be at most one NODE_HANDLE message in every chunk we
         // read because SCM_RIGHTS messages don't get coalesced. Make sure
         // that we deliver the handle with the right message however.
-        if (message && message.cmd === 'NODE_HANDLE')
-          handleMessage(target, message, recvHandle);
-        else
-          handleMessage(target, message, undefined);
+        if (isInternal(message)) {
+          if (message.cmd === 'NODE_HANDLE') {
+            handleMessage(message, pendingHandle, true);
+            pendingHandle = null;
+          } else {
+            handleMessage(message, undefined, true);
+          }
+        } else {
+          handleMessage(message, undefined, false);
+        }
       }
       jsonBuffer = incompleteChunk;
       this.buffering = jsonBuffer.length !== 0;
@@ -478,7 +531,7 @@ function setupChannel(target, channel) {
       target.disconnect();
       channel.onread = nop;
       channel.close();
-      target._channel = null;
+      target.channel = null;
       maybeClose(target);
     }
   };
@@ -489,23 +542,38 @@ function setupChannel(target, channel) {
   // handlers will go through this
   target.on('internalMessage', function(message, handle) {
     // Once acknowledged - continue sending handles.
-    if (message.cmd === 'NODE_HANDLE_ACK') {
-      if (target._pendingHandle) {
-        target._pendingHandle.close();
-        target._pendingHandle = null;
+    if (message.cmd === 'NODE_HANDLE_ACK' ||
+        message.cmd === 'NODE_HANDLE_NACK') {
+
+      if (target._pendingMessage) {
+        if (message.cmd === 'NODE_HANDLE_ACK') {
+          closePendingHandle(target);
+        } else if (target._pendingMessage.retransmissions++ ===
+                   MAX_HANDLE_RETRANSMISSIONS) {
+          closePendingHandle(target);
+          process.emitWarning('Handle did not reach the receiving process ' +
+                              'correctly', 'SentHandleNotReceivedWarning');
+        }
       }
 
       assert(Array.isArray(target._handleQueue));
       var queue = target._handleQueue;
       target._handleQueue = null;
 
+      if (target._pendingMessage) {
+        target._send(target._pendingMessage.message,
+                     target._pendingMessage.handle,
+                     target._pendingMessage.options,
+                     target._pendingMessage.callback);
+      }
+
       for (var i = 0; i < queue.length; i++) {
         var args = queue[i];
         target._send(args.message, args.handle, args.options, args.callback);
       }
 
       // Process a pending disconnect (if any).
-      if (!target.connected && target._channel && !target._handleQueue)
+      if (!target.connected && target.channel && !target._handleQueue)
         target._disconnect();
 
       return;
@@ -513,6 +581,12 @@ function setupChannel(target, channel) {
 
     if (message.cmd !== 'NODE_HANDLE') return;
 
+    // It is possible that the handle is not received because of some error on
+    // ancillary data reception such as MSG_CTRUNC. In this case, report the
+    // sender about it by sending a NODE_HANDLE_NACK message.
+    if (!handle)
+      return target._send({ cmd: 'NODE_HANDLE_NACK' }, null, true);
+
     // Acknowledge handle receival. Don't emit error events (for example if
     // the other side has disconnected) because this call to send() is not
     // initiated by the user and it shouldn't be fatal to be unable to ACK
@@ -529,7 +603,7 @@ function setupChannel(target, channel) {
 
     // Convert handle object
     obj.got.call(this, message, handle, function(handle) {
-      handleMessage(target, message.msg, handle);
+      handleMessage(message.msg, handle, isInternal(message.msg));
     });
   });
 
@@ -543,32 +617,32 @@ function setupChannel(target, channel) {
       options = undefined;
     } else if (options !== undefined &&
                (options === null || typeof options !== 'object')) {
-      throw new TypeError('"options" argument must be an object');
+      throw new ERR_INVALID_ARG_TYPE('options', 'Object', options);
     }
 
-    options = Object.assign({swallowErrors: false}, options);
+    options = Object.assign({ swallowErrors: false }, options);
 
     if (this.connected) {
       return this._send(message, handle, options, callback);
     }
-    const ex = new Error('channel closed');
+    const ex = new ERR_IPC_CHANNEL_CLOSED();
     if (typeof callback === 'function') {
       process.nextTick(callback, ex);
     } else {
-      this.emit('error', ex);  // FIXME(bnoordhuis) Defer to next tick.
+      process.nextTick(() => this.emit('error', ex));
     }
     return false;
   };
 
   target._send = function(message, handle, options, callback) {
-    assert(this.connected || this._channel);
+    assert(this.connected || this.channel);
 
     if (message === undefined)
-      throw new TypeError('"message" argument cannot be undefined');
+      throw new ERR_MISSING_ARGS('message');
 
     // Support legacy function signature
     if (typeof options === 'boolean') {
-      options = {swallowErrors: options};
+      options = { swallowErrors: options };
     }
 
     // package messages with a handle object
@@ -591,7 +665,7 @@ function setupChannel(target, channel) {
       } else if (handle instanceof UDP) {
         message.type = 'dgram.Native';
       } else {
-        throw new TypeError('This handle type can\'t be sent');
+        throw new ERR_INVALID_HANDLE_TYPE();
       }
 
       // Queue-up message and handle if we haven't received ACK yet.
@@ -623,7 +697,8 @@ function setupChannel(target, channel) {
         net._setSimultaneousAccepts(handle);
       }
     } else if (this._handleQueue &&
-               !(message && message.cmd === 'NODE_HANDLE_ACK')) {
+               !(message && (message.cmd === 'NODE_HANDLE_ACK' ||
+                             message.cmd === 'NODE_HANDLE_NACK'))) {
       // Queue request anyway to avoid out-of-order messages.
       this._handleQueue.push({
         callback: callback,
@@ -645,31 +720,30 @@ function setupChannel(target, channel) {
         if (!this._handleQueue)
           this._handleQueue = [];
         if (obj && obj.postSend)
-          obj.postSend(handle, options, target);
+          obj.postSend(message, handle, options, callback, target);
       }
 
-      req.oncomplete = function() {
-        if (this.async === true)
+      if (req.async) {
+        req.oncomplete = function() {
           control.unref();
-        if (typeof callback === 'function')
-          callback(null);
-      };
-      if (req.async === true) {
+          if (typeof callback === 'function')
+            callback(null);
+        };
         control.ref();
-      } else {
-        process.nextTick(function() { req.oncomplete(); });
+      } else if (typeof callback === 'function') {
+        process.nextTick(callback, null);
       }
     } else {
       // Cleanup handle on error
       if (obj && obj.postSend)
-        obj.postSend(handle, options);
+        obj.postSend(message, handle, options, callback);
 
       if (!options.swallowErrors) {
         const ex = errnoException(err, 'write');
         if (typeof callback === 'function') {
           process.nextTick(callback, ex);
         } else {
-          this.emit('error', ex);  // FIXME(bnoordhuis) Defer to next tick.
+          process.nextTick(() => this.emit('error', ex));
         }
       }
     }
@@ -681,17 +755,17 @@ function setupChannel(target, channel) {
   // connected will be set to false immediately when a disconnect() is
   // requested, even though the channel might still be alive internally to
   // process queued messages. The three states are distinguished as follows:
-  // - disconnect() never requested: _channel is not null and connected
+  // - disconnect() never requested: channel is not null and connected
   //   is true
-  // - disconnect() requested, messages in the queue: _channel is not null
+  // - disconnect() requested, messages in the queue: channel is not null
   //   and connected is false
-  // - disconnect() requested, channel actually disconnected: _channel is
+  // - disconnect() requested, channel actually disconnected: channel is
   //   null and connected is false
   target.connected = true;
 
   target.disconnect = function() {
     if (!this.connected) {
-      this.emit('error', new Error('IPC channel is already disconnected'));
+      this.emit('error', new ERR_IPC_DISCONNECTED());
       return;
     }
 
@@ -706,15 +780,13 @@ function setupChannel(target, channel) {
   };
 
   target._disconnect = function() {
-    assert(this._channel);
+    assert(this.channel);
 
     // This marks the fact that the channel is actually disconnected.
-    this._channel = null;
+    this.channel = null;
 
-    if (this._pendingHandle) {
-      this._pendingHandle.close();
-      this._pendingHandle = null;
-    }
+    if (this._pendingMessage)
+      closePendingHandle(this);
 
     var fired = false;
     function finish() {
@@ -736,27 +808,30 @@ function setupChannel(target, channel) {
     process.nextTick(finish);
   };
 
+  function emit(event, message, handle) {
+    target.emit(event, message, handle);
+  }
+
+  function handleMessage(message, handle, internal) {
+    if (!target.channel)
+      return;
+
+    var eventName = (internal ? 'internalMessage' : 'message');
+
+    process.nextTick(emit, eventName, message, handle);
+  }
+
   channel.readStart();
   return control;
 }
 
-
 const INTERNAL_PREFIX = 'NODE_';
-function handleMessage(target, message, handle) {
-  if (!target._channel)
-    return;
-
-  var eventName = 'message';
-  if (message !== null &&
-      typeof message === 'object' &&
-      typeof message.cmd === 'string' &&
-      message.cmd.length > INTERNAL_PREFIX.length &&
-      message.cmd.slice(0, INTERNAL_PREFIX.length) === INTERNAL_PREFIX) {
-    eventName = 'internalMessage';
-  }
-  process.nextTick(() => {
-    target.emit(eventName, message, handle);
-  });
+function isInternal(message) {
+  return (message !== null &&
+          typeof message === 'object' &&
+          typeof message.cmd === 'string' &&
+          message.cmd.length > INTERNAL_PREFIX.length &&
+          message.cmd.slice(0, INTERNAL_PREFIX.length) === INTERNAL_PREFIX);
 }
 
 function nop() { }
@@ -771,11 +846,11 @@ function _validateStdio(stdio, sync) {
       case 'ignore': stdio = ['ignore', 'ignore', 'ignore']; break;
       case 'pipe': stdio = ['pipe', 'pipe', 'pipe']; break;
       case 'inherit': stdio = [0, 1, 2]; break;
-      default: throw new TypeError('Incorrect value of stdio option: ' + stdio);
+      default:
+        throw new ERR_INVALID_OPT_VALUE('stdio', stdio);
     }
   } else if (!Array.isArray(stdio)) {
-    throw new TypeError('Incorrect value of stdio option: ' +
-        util.inspect(stdio));
+    throw new ERR_INVALID_OPT_VALUE('stdio', util.inspect(stdio));
   }
 
   // At least 3 stdio will be created
@@ -788,12 +863,10 @@ function _validateStdio(stdio, sync) {
   // (i.e. PipeWraps or fds)
   stdio = stdio.reduce(function(acc, stdio, i) {
     function cleanup() {
-      acc.filter(function(stdio) {
-        return stdio.type === 'pipe' || stdio.type === 'ipc';
-      }).forEach(function(stdio) {
-        if (stdio.handle)
-          stdio.handle.close();
-      });
+      for (var i = 0; i < acc.length; i++) {
+        if ((acc[i].type === 'pipe' || acc[i].type === 'ipc') && acc[i].handle)
+          acc[i].handle.close();
+      }
     }
 
     // Defaults
@@ -802,7 +875,7 @@ function _validateStdio(stdio, sync) {
     }
 
     if (stdio === 'ignore') {
-      acc.push({type: 'ignore'});
+      acc.push({ type: 'ignore' });
     } else if (stdio === 'pipe' || typeof stdio === 'number' && stdio < 0) {
       var a = {
         type: 'pipe',
@@ -811,7 +884,7 @@ function _validateStdio(stdio, sync) {
       };
 
       if (!sync)
-        a.handle = new Pipe();
+        a.handle = new Pipe(PipeConstants.SOCKET);
 
       acc.push(a);
     } else if (stdio === 'ipc') {
@@ -819,12 +892,12 @@ function _validateStdio(stdio, sync) {
         // Cleanup previously created pipes
         cleanup();
         if (!sync)
-          throw new Error('Child process can have only one IPC pipe');
+          throw new ERR_IPC_ONE_PIPE();
         else
-          throw new Error('You cannot use IPC with synchronous forks');
+          throw new ERR_IPC_SYNC_FORK();
       }
 
-      ipc = new Pipe(true);
+      ipc = new Pipe(PipeConstants.IPC);
       ipcFd = i;
 
       acc.push({
@@ -845,40 +918,38 @@ function _validateStdio(stdio, sync) {
     } else if (getHandleWrapType(stdio) || getHandleWrapType(stdio.handle) ||
                getHandleWrapType(stdio._handle)) {
       var handle = getHandleWrapType(stdio) ?
-          stdio :
-          getHandleWrapType(stdio.handle) ? stdio.handle : stdio._handle;
+        stdio :
+        getHandleWrapType(stdio.handle) ? stdio.handle : stdio._handle;
 
       acc.push({
         type: 'wrap',
         wrapType: getHandleWrapType(handle),
         handle: handle
       });
-    } else if (stdio instanceof Buffer || typeof stdio === 'string') {
+    } else if (isArrayBufferView(stdio) || typeof stdio === 'string') {
       if (!sync) {
         cleanup();
-        throw new TypeError('Asynchronous forks do not support Buffer input: ' +
-            util.inspect(stdio));
+        throw new ERR_INVALID_SYNC_FORK_INPUT(util.inspect(stdio));
       }
     } else {
       // Cleanup
       cleanup();
-      throw new TypeError('Incorrect value for stdio stream: ' +
-          util.inspect(stdio));
+      throw new ERR_INVALID_OPT_VALUE('stdio', util.inspect(stdio));
     }
 
     return acc;
   }, []);
 
-  return {stdio: stdio, ipc: ipc, ipcFd: ipcFd};
+  return { stdio, ipc, ipcFd };
 }
 
 
-function getSocketList(type, slave, key) {
-  var sockets = slave._channel.sockets[type];
+function getSocketList(type, worker, key) {
+  var sockets = worker.channel.sockets[type];
   var socketList = sockets[key];
   if (!socketList) {
     var Construct = type === 'send' ? SocketListSend : SocketListReceive;
-    socketList = sockets[key] = new Construct(slave, key);
+    socketList = sockets[key] = new Construct(worker, key);
   }
   return socketList;
 }
@@ -891,3 +962,34 @@ function maybeClose(subprocess) {
     subprocess.emit('close', subprocess.exitCode, subprocess.signalCode);
   }
 }
+
+function spawnSync(opts) {
+  var options = opts.options;
+  var result = spawn_sync.spawn(options);
+
+  if (result.output && options.encoding && options.encoding !== 'buffer') {
+    for (var i = 0; i < result.output.length; i++) {
+      if (!result.output[i])
+        continue;
+      result.output[i] = result.output[i].toString(options.encoding);
+    }
+  }
+
+  result.stdout = result.output && result.output[1];
+  result.stderr = result.output && result.output[2];
+
+  if (result.error) {
+    result.error = errnoException(result.error, 'spawnSync ' + opts.file);
+    result.error.path = opts.file;
+    result.error.spawnargs = opts.args.slice(1);
+  }
+
+  return result;
+}
+
+module.exports = {
+  ChildProcess,
+  setupChannel,
+  _validateStdio,
+  spawnSync
+};
