@@ -1,15 +1,15 @@
 'use strict';
 
 const common = require('../common');
-const assert = require('assert');
-const path = require('path');
-const childProcess = require('child_process');
-
+const fixtures = require('../common/fixtures');
 // Refs: https://github.com/nodejs/node/pull/2253
-if (common.isSunOS) {
+if (common.isSunOS)
   common.skip('unreliable on SunOS');
-  return;
-}
+if (!common.isMainThread)
+  common.skip('process.chdir is not available in Workers');
+
+const assert = require('assert');
+const childProcess = require('child_process');
 
 const nodeBinary = process.argv[0];
 
@@ -21,18 +21,16 @@ const preloadOption = (preloads) => {
   return option;
 };
 
-const fixture = (name) => path.join(common.fixturesDir, name);
-
-const fixtureA = fixture('printA.js');
-const fixtureB = fixture('printB.js');
-const fixtureC = fixture('printC.js');
-const fixtureD = fixture('define-global.js');
-const fixtureThrows = fixture('throws_error4.js');
+const fixtureA = fixtures.path('printA.js');
+const fixtureB = fixtures.path('printB.js');
+const fixtureC = fixtures.path('printC.js');
+const fixtureD = fixtures.path('define-global.js');
+const fixtureThrows = fixtures.path('throws_error4.js');
 
 // test preloading a single module works
 childProcess.exec(`"${nodeBinary}" ${preloadOption([fixtureA])} "${fixtureB}"`,
                   function(err, stdout, stderr) {
-                    if (err) throw err;
+                    assert.ifError(err);
                     assert.strictEqual(stdout, 'A\nB\n');
                   });
 
@@ -40,7 +38,7 @@ childProcess.exec(`"${nodeBinary}" ${preloadOption([fixtureA])} "${fixtureB}"`,
 childProcess.exec(
   `"${nodeBinary}" ${preloadOption([fixtureA, fixtureB])} "${fixtureC}"`,
   function(err, stdout, stderr) {
-    if (err) throw err;
+    assert.ifError(err);
     assert.strictEqual(stdout, 'A\nB\nC\n');
   }
 );
@@ -61,7 +59,7 @@ childProcess.exec(
 childProcess.exec(
   `"${nodeBinary}" ${preloadOption([fixtureA])}-e "console.log('hello');"`,
   function(err, stdout, stderr) {
-    if (err) throw err;
+    assert.ifError(err);
     assert.strictEqual(stdout, 'A\nhello\n');
   }
 );
@@ -70,7 +68,7 @@ childProcess.exec(
 const stdinProc = childProcess.spawn(
   nodeBinary,
   ['--require', fixtureA],
-  {stdio: 'pipe'}
+  { stdio: 'pipe' }
 );
 stdinProc.stdin.end("console.log('hello');");
 let stdinStdout = '';
@@ -86,7 +84,7 @@ stdinProc.on('close', function(code) {
 const replProc = childProcess.spawn(
   nodeBinary,
   ['-i', '--require', fixtureA],
-  {stdio: 'pipe'}
+  { stdio: 'pipe' }
 );
 replProc.stdin.end('.exit\n');
 let replStdout = '';
@@ -106,9 +104,9 @@ replProc.on('close', function(code) {
 // also test that duplicated preload only gets loaded once
 childProcess.exec(
   `"${nodeBinary}" ${preloadOption([fixtureA])}-e "console.log('hello');" ${
-  preloadOption([fixtureA, fixtureB])}`,
+    preloadOption([fixtureA, fixtureB])}`,
   function(err, stdout, stderr) {
-    if (err) throw err;
+    assert.ifError(err);
     assert.strictEqual(stdout, 'A\nB\nhello\n');
   }
 );
@@ -126,21 +124,21 @@ interactive.stdin.write('a\n');
 interactive.stdin.write('process.exit()\n');
 
 childProcess.exec(
-  `"${nodeBinary}" --require "${fixture('cluster-preload.js')}" "${
-  fixture('cluster-preload-test.js')}"`,
+  `"${nodeBinary}" --require "${fixtures.path('cluster-preload.js')}" "${
+    fixtures.path('cluster-preload-test.js')}"`,
   function(err, stdout, stderr) {
-    if (err) throw err;
+    assert.ifError(err);
     assert.ok(/worker terminated with code 43/.test(stdout));
   }
 );
 
 // https://github.com/nodejs/node/issues/1691
-process.chdir(common.fixturesDir);
+process.chdir(fixtures.fixturesDir);
 childProcess.exec(
   `"${nodeBinary}" --expose_natives_as=v8natives --require ` +
-     `"${fixture('cluster-preload.js')}" cluster-preload-test.js`,
+     `"${fixtures.path('cluster-preload.js')}" cluster-preload-test.js`,
   function(err, stdout, stderr) {
-    if (err) throw err;
+    assert.ifError(err);
     assert.ok(/worker terminated with code 43/.test(stdout));
   }
 );
