@@ -3,6 +3,8 @@ const common = require('../common');
 const assert = require('assert');
 const domain = require('domain');
 
+common.disableCrashOnUnhandledRejection();
+
 const asyncTest = (function() {
   let asyncTestsEnabled = false;
   let asyncTestLastCheck;
@@ -11,9 +13,9 @@ const asyncTest = (function() {
   let currentTest = null;
 
   function fail(error) {
-    const stack = currentTest
-          ? `${error.stack}\nFrom previous event:\n${currentTest.stack}`
-          : error.stack;
+    const stack = currentTest ?
+      `${error.stack}\nFrom previous event:\n${currentTest.stack}` :
+      error.stack;
 
     if (currentTest)
       process.stderr.write(`'${currentTest.description}' failed\n\n`);
@@ -110,7 +112,7 @@ asyncTest('synchronously rejected promise should trigger' +
           ' unhandledRejection', function(done) {
   const e = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e, reason);
+    assert.strictEqual(reason, e);
   });
   Promise.reject(e);
 });
@@ -119,7 +121,7 @@ asyncTest('synchronously rejected promise should trigger' +
           ' unhandledRejection', function(done) {
   const e = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e, reason);
+    assert.strictEqual(reason, e);
   });
   new Promise(function(_, reject) {
     reject(e);
@@ -130,7 +132,7 @@ asyncTest('Promise rejected after setImmediate should trigger' +
           ' unhandledRejection', function(done) {
   const e = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e, reason);
+    assert.strictEqual(reason, e);
   });
   new Promise(function(_, reject) {
     setImmediate(function() {
@@ -143,7 +145,7 @@ asyncTest('Promise rejected after setTimeout(,1) should trigger' +
           ' unhandled rejection', function(done) {
   const e = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e, reason);
+    assert.strictEqual(reason, e);
   });
   new Promise(function(_, reject) {
     setTimeout(function() {
@@ -156,7 +158,7 @@ asyncTest('Catching a promise rejection after setImmediate is not' +
           ' soon enough to stop unhandledRejection', function(done) {
   const e = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e, reason);
+    assert.strictEqual(reason, e);
   });
   let _reject;
   const promise = new Promise(function(_, reject) {
@@ -164,7 +166,7 @@ asyncTest('Catching a promise rejection after setImmediate is not' +
   });
   _reject(e);
   setImmediate(function() {
-    promise.then(common.fail, function() {});
+    promise.then(assert.fail, function() {});
   });
 });
 
@@ -173,20 +175,20 @@ asyncTest('When re-throwing new errors in a promise catch, only the' +
   const e = new Error();
   const e2 = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e2, reason);
-    assert.strictEqual(promise2, promise);
+    assert.strictEqual(reason, e2);
+    assert.strictEqual(promise, promise2);
   });
-  let promise2 = Promise.reject(e).then(common.fail, function(reason) {
-    assert.strictEqual(e, reason);
+  const promise2 = Promise.reject(e).then(assert.fail, function(reason) {
+    assert.strictEqual(reason, e);
     throw e2;
   });
 });
 
-asyncTest('Test params of unhandledRejection for a synchronously-rejected' +
+asyncTest('Test params of unhandledRejection for a synchronously-rejected ' +
           'promise', function(done) {
   const e = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e, reason);
+    assert.strictEqual(reason, e);
     assert.strictEqual(promise, promise);
   });
   Promise.reject(e);
@@ -198,15 +200,15 @@ asyncTest('When re-throwing new errors in a promise catch, only the ' +
   const e = new Error();
   const e2 = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e2, reason);
-    assert.strictEqual(promise2, promise);
+    assert.strictEqual(reason, e2);
+    assert.strictEqual(promise, promise2);
   });
-  let promise2 = new Promise(function(_, reject) {
+  const promise2 = new Promise(function(_, reject) {
     setTimeout(function() {
       reject(e);
     }, 1);
-  }).then(common.fail, function(reason) {
-    assert.strictEqual(e, reason);
+  }).then(assert.fail, function(reason) {
+    assert.strictEqual(reason, e);
     throw e2;
   });
 });
@@ -217,15 +219,15 @@ asyncTest('When re-throwing new errors in a promise catch, only the re-thrown' +
   const e = new Error();
   const e2 = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e2, reason);
-    assert.strictEqual(promise2, promise);
+    assert.strictEqual(reason, e2);
+    assert.strictEqual(promise, promise2);
   });
   const promise = new Promise(function(_, reject) {
     setTimeout(function() {
       reject(e);
       process.nextTick(function() {
-        promise2 = promise.then(common.fail, function(reason) {
-          assert.strictEqual(e, reason);
+        promise2 = promise.then(assert.fail, function(reason) {
+          assert.strictEqual(reason, e);
           throw e2;
         });
       });
@@ -240,7 +242,7 @@ asyncTest(
   function(done) {
     const e = new Error();
     onUnhandledFail(done);
-    Promise.reject(e).then(common.fail, function() {});
+    Promise.reject(e).then(assert.fail, function() {});
   }
 );
 
@@ -252,7 +254,7 @@ asyncTest(
     onUnhandledFail(done);
     new Promise(function(_, reject) {
       reject(e);
-    }).then(common.fail, function() {});
+    }).then(assert.fail, function() {});
   }
 );
 
@@ -262,7 +264,7 @@ asyncTest('Attaching a promise catch in a process.nextTick is soon enough to' +
   onUnhandledFail(done);
   const promise = Promise.reject(e);
   process.nextTick(function() {
-    promise.then(common.fail, function() {});
+    promise.then(assert.fail, function() {});
   });
 });
 
@@ -274,7 +276,7 @@ asyncTest('Attaching a promise catch in a process.nextTick is soon enough to' +
     reject(e);
   });
   process.nextTick(function() {
-    promise.then(common.fail, function() {});
+    promise.then(assert.fail, function() {});
   });
 });
 
@@ -293,8 +295,8 @@ asyncTest('While inside setImmediate, catching a rejected promise derived ' +
   });
 });
 
-// State adapation tests
-asyncTest('catching a promise which is asynchronously rejected (via' +
+// State adaptation tests
+asyncTest('catching a promise which is asynchronously rejected (via ' +
           'resolution to an asynchronously-rejected promise) prevents' +
           ' unhandledRejection', function(done) {
   const e = new Error();
@@ -305,8 +307,8 @@ asyncTest('catching a promise which is asynchronously rejected (via' +
         reject(e);
       }, 1);
     });
-  }).then(common.fail, function(reason) {
-    assert.strictEqual(e, reason);
+  }).then(assert.fail, function(reason) {
+    assert.strictEqual(reason, e);
   });
 });
 
@@ -316,8 +318,8 @@ asyncTest('Catching a rejected promise derived from throwing in a' +
   onUnhandledFail(done);
   Promise.resolve().then(function() {
     throw e;
-  }).then(common.fail, function(reason) {
-    assert.strictEqual(e, reason);
+  }).then(assert.fail, function(reason) {
+    assert.strictEqual(reason, e);
   });
 });
 
@@ -328,8 +330,8 @@ asyncTest('Catching a rejected promise derived from returning a' +
   onUnhandledFail(done);
   Promise.resolve().then(function() {
     return Promise.reject(e);
-  }).then(common.fail, function(reason) {
-    assert.strictEqual(e, reason);
+  }).then(assert.fail, function(reason) {
+    assert.strictEqual(reason, e);
   });
 });
 
@@ -337,12 +339,11 @@ asyncTest('A rejected promise derived from returning an' +
           ' asynchronously-rejected promise in a fulfillment handler' +
           ' does trigger unhandledRejection', function(done) {
   const e = new Error();
-  let _promise;
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e, reason);
-    assert.strictEqual(_promise, promise);
+    assert.strictEqual(reason, e);
+    assert.strictEqual(promise, _promise);
   });
-  _promise = Promise.resolve().then(function() {
+  const _promise = Promise.resolve().then(function() {
     return new Promise(function(_, reject) {
       setTimeout(function() {
         reject(e);
@@ -354,12 +355,11 @@ asyncTest('A rejected promise derived from returning an' +
 asyncTest('A rejected promise derived from throwing in a fulfillment handler' +
           ' does trigger unhandledRejection', function(done) {
   const e = new Error();
-  let _promise;
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e, reason);
-    assert.strictEqual(_promise, promise);
+    assert.strictEqual(reason, e);
+    assert.strictEqual(promise, _promise);
   });
-  _promise = Promise.resolve().then(function() {
+  const _promise = Promise.resolve().then(function() {
     throw e;
   });
 });
@@ -369,23 +369,22 @@ asyncTest(
   ' promise in a fulfillment handler does trigger unhandledRejection',
   function(done) {
     const e = new Error();
-    let _promise;
     onUnhandledSucceed(done, function(reason, promise) {
-      assert.strictEqual(e, reason);
-      assert.strictEqual(_promise, promise);
+      assert.strictEqual(reason, e);
+      assert.strictEqual(promise, _promise);
     });
-    _promise = Promise.resolve().then(function() {
+    const _promise = Promise.resolve().then(function() {
       return Promise.reject(e);
     });
   }
 );
 
 // Combinations with Promise.all
-asyncTest('Catching the Promise.all() of a collection that includes a' +
+asyncTest('Catching the Promise.all() of a collection that includes a ' +
           'rejected promise prevents unhandledRejection', function(done) {
   const e = new Error();
   onUnhandledFail(done);
-  Promise.all([Promise.reject(e)]).then(common.fail, function() {});
+  Promise.all([Promise.reject(e)]).then(assert.fail, function() {});
 });
 
 asyncTest(
@@ -401,7 +400,7 @@ asyncTest(
     });
     p = Promise.all([p]);
     process.nextTick(function() {
-      p.then(common.fail, function() {});
+      p.then(assert.fail, function() {});
     });
   }
 );
@@ -411,10 +410,10 @@ asyncTest('Failing to catch the Promise.all() of a collection that includes' +
           ' promise, not the passed promise', function(done) {
   const e = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e, reason);
-    assert.strictEqual(p, promise);
+    assert.strictEqual(reason, e);
+    assert.strictEqual(promise, p);
   });
-  let p = Promise.all([Promise.reject(e)]);
+  const p = Promise.all([Promise.reject(e)]);
 });
 
 asyncTest('Waiting setTimeout(, 10) to catch a promise causes an' +
@@ -423,22 +422,22 @@ asyncTest('Waiting setTimeout(, 10) to catch a promise causes an' +
   const unhandledPromises = [];
   const e = new Error();
   process.on('unhandledRejection', function(reason, promise) {
-    assert.strictEqual(e, reason);
+    assert.strictEqual(reason, e);
     unhandledPromises.push(promise);
   });
   process.on('rejectionHandled', function(promise) {
-    assert.strictEqual(1, unhandledPromises.length);
+    assert.strictEqual(unhandledPromises.length, 1);
     assert.strictEqual(unhandledPromises[0], promise);
-    assert.strictEqual(thePromise, promise);
+    assert.strictEqual(promise, thePromise);
     done();
   });
 
-  let thePromise = new Promise(function() {
+  const thePromise = new Promise(function() {
     throw e;
   });
   setTimeout(function() {
-    thePromise.then(common.fail, function(reason) {
-      assert.strictEqual(e, reason);
+    thePromise.then(assert.fail, function(reason) {
+      assert.strictEqual(reason, e);
     });
   }, 10);
 });
@@ -569,10 +568,10 @@ asyncTest('setImmediate + promise microtasks is too late to attach a catch' +
           ' (setImmediate before promise creation/rejection)', function(done) {
   const e = new Error();
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(e, reason);
-    assert.strictEqual(p, promise);
+    assert.strictEqual(reason, e);
+    assert.strictEqual(promise, p);
   });
-  let p = Promise.reject(e);
+  const p = Promise.reject(e);
   setImmediate(function() {
     Promise.resolve().then(function() {
       p.catch(function() {});
@@ -584,8 +583,8 @@ asyncTest('setImmediate + promise microtasks is too late to attach a catch' +
           ' handler; unhandledRejection will be triggered in that case' +
           ' (setImmediate before promise creation/rejection)', function(done) {
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(undefined, reason);
-    assert.strictEqual(p, promise);
+    assert.strictEqual(reason, undefined);
+    assert.strictEqual(promise, p);
   });
   setImmediate(function() {
     Promise.resolve().then(function() {
@@ -598,17 +597,17 @@ asyncTest('setImmediate + promise microtasks is too late to attach a catch' +
       });
     });
   });
-  let p = Promise.reject();
+  const p = Promise.reject();
 });
 
 asyncTest('setImmediate + promise microtasks is too late to attach a catch' +
           ' handler; unhandledRejection will be triggered in that case' +
           ' (setImmediate after promise creation/rejection)', function(done) {
   onUnhandledSucceed(done, function(reason, promise) {
-    assert.strictEqual(undefined, reason);
-    assert.strictEqual(p, promise);
+    assert.strictEqual(reason, undefined);
+    assert.strictEqual(promise, p);
   });
-  let p = Promise.reject();
+  const p = Promise.reject();
   setImmediate(function() {
     Promise.resolve().then(function() {
       Promise.resolve().then(function() {
@@ -635,9 +634,8 @@ asyncTest(
       const e = new Error('error');
       const domainError = new Error('domain error');
       onUnhandledSucceed(done, function(reason, promise) {
-        assert.strictEqual(reason, e);
-        assert.strictEqual(domainReceivedError, domainError);
-        d.dispose();
+        assert.strictEqual(e, reason);
+        assert.strictEqual(domainError, domainReceivedError);
       });
       Promise.reject(e);
       process.nextTick(function() {
@@ -666,13 +664,13 @@ asyncTest('nextTick is immediately scheduled when called inside an event' +
 });
 
 asyncTest('Throwing an error inside a rejectionHandled handler goes to' +
-          ' unhandledException, and does not cause .catch() to throw an' +
+          ' unhandledException, and does not cause .catch() to throw an ' +
           'exception', function(done) {
   clean();
   const e = new Error();
   const e2 = new Error();
   const tearDownException = setupException(function(err) {
-    assert.strictEqual(e2, err);
+    assert.strictEqual(err, e2);
     tearDownException();
     done();
   });
@@ -683,8 +681,21 @@ asyncTest('Throwing an error inside a rejectionHandled handler goes to' +
   setTimeout(function() {
     try {
       p.catch(function() {});
-    } catch (e) {
+    } catch {
       done(new Error('fail'));
     }
   }, 1);
 });
+
+asyncTest('Rejected promise inside unhandledRejection allows nextTick loop' +
+          ' to proceed first', function(done) {
+  clean();
+  Promise.reject(0);
+  let didCall = false;
+  process.on('unhandledRejection', () => {
+    assert(!didCall);
+    didCall = true;
+    const promise = Promise.reject(0);
+    process.nextTick(() => promise.catch(() => done()));
+  });
+});
